#
#	A help tool in TCL
#
#
#

option add *TclHelp*text.height 24 widgetDefault
option add *TclHelp*text.width 80 widgetDefault

proc helpWindow { w { parent . }} {

    if ![winfo exists $w] {
	toplevel $w -class TclHelp
	wm withdraw $w
	wm resizable $w yes yes
	
	# header with Close button:
	frame $w.buttonframe
	pack $w.buttonframe -side top -fill x -expand no -pady 2m
	button $w.buttonframe.close -text "Close" -command "wm iconify $w"
	pack   $w.buttonframe.close -expand no -side left -padx 10

	# scrolled text:
	frame $w.text -borderwidth 1
	pack $w.text -fill both -expand yes		
	scrollbar $w.text.vscroll -command "$w.text.text yview"
	pack      $w.text.vscroll -side right -fill y
	text $w.text.text -yscrollcommand "$w.text.vscroll set"
	pack $w.text.text -side left -fill both -expand yes

	# tags
	set tw $w.text.text
	$tw config -wrap word
	set hfont helvetica
	set fsize 11
	$tw tag configure TI -font [list $hfont 18 bold] -spacing3 12 
	$tw tag configure STI -font [list $hfont 18 bold] -spacing3 12
	$tw tag configure H1 -font [list $hfont 14 bold] -spacing1 4 -spacing3 4 -underline yes
	$tw tag configure H2 -font [list $hfont 12 bold] -spacing1 4 -spacing3 4
	$tw tag configure H3 -font [list $hfont $fsize bold] -spacing1 4 -spacing3 4
	$tw tag configure BB -font [list $hfont $fsize bold] -lmargin1 20 -lmargin2 20
	$tw tag configure BU -font [list $hfont 6 bold] -lmargin1 20 -lmargin2 20
	$tw tag configure P -font [list $hfont $fsize] -lmargin1 $fsize
	$tw tag configure PS -font [list $hfont $fsize] -lmargin1 $fsize
	$tw tag configure T -font [list $hfont $fsize]
	$tw tag configure B -font [list $hfont $fsize bold]
	$tw tag configure I -font [list $hfont $fsize italic]
	$tw tag configure BI -font [list $hfont $fsize bold italic]
	$tw tag configure REF -font [list $hfont $fsize bold] -foreground darkblue -underline yes 
	$tw tag configure mysel -foreground "#4090ff"
	bind $tw <1> {}
	bind $tw <B1-Motion> {}
	$tw tag bind REF <1> "hyperJump $tw %x %y"
	$tw tag configure ANCHOR

	# place it once
	set wp [winfo reqwidth $parent]
	set hp [winfo reqheight $parent]
	set x [expr [winfo rootx $parent] + ( $wp - [winfo reqwidth $w] ) / 2 ]
	set y [expr [winfo rooty $parent] + ( $hp - [winfo reqheight $w] ) / 2 ]
	wm geom $w +$x+$y
    }

    # Show the view:
    raise $w
    update idletasks
    wm deiconify $w
}

set BlockSTags { H1 H2 H3 P }
set BlockETags { ETI }

proc helpWindowOpen { w filename { title {}} } {
    global BlockETags BlockSTags Anchors References
    if {[catch "open {$filename} r" stream]} {
	tk_dialog .msg "Help Error:" "$filename not accessible" error 0 OK
	return
    }
    wm title $w $title

    set tw $w.text.text
    $tw configure -state normal
    $tw delete 0.0 end
    while { [gets $stream text] >= 0 } {
	set tagend [string wordend $text 0]
	set tag [string range $text 0 [expr $tagend - 1]]
	incr tagend

	if { [lsearch -exact $BlockSTags $tag] >= 0 } {
	    $tw insert end "\n"
	}
	switch -exact $tag {
	    BU {
		$tw insert end "o  " $tag
	    }
	    ANCHOR {
		set Anchors([string range $text $tagend end]) [$tw index end]
		continue
	    }
	    REF {
		#puts stdout "REF at [$tw index end-1char]" 
		set labelend [string wordend $text $tagend]
		set References([$tw index end-1char]) \
			[string range $text $tagend [expr $labelend - 1]]
		set tagend [expr $labelend + 1]
	    }
	}
	$w.text.text insert end [string range $text $tagend end] $tag

	if { [lsearch -exact $BlockETags $tag] >= 0 } {
	    $tw insert end "\n"
	}
    }
    close $stream
    #$w.text.text configure -state disabled
}

proc hyperJump { win x y } {

    global References Anchors

    set index [$win index "@$x,$y"]
    set range [$win tag prevrange REF $index]
    update idletasks
    update
    set err ""
    catch {
	set label $References([lindex $range 0])
	set loc $Anchors($label)
	#puts stdout "jump to $loc"
	$win configure -state normal
	$win see $loc
	$win tag remove mysel 0.0 end
        $win tag add mysel "$loc linestart " "$loc lineend"
	$win configure -state disabled
    } err
    if { $err != "" } {
	bell
    }
}