//-----------------------------------------------------------------------
//
//	Project TibThob: a tool to search the Rangjung Yeshe Dictionary
//
//
//	This program is subject to the GNU GENERAL PUBLIC LICENSE
//	It is provided as is - WITHOUT ANY WARRANTY - in the hope
//	it will be of benefit to all beings, or at least some of them.
//	
//-----------------------------------------------------------------------
//	Copyright Xavier C. FRANC 1999
//	report bugs and comments at:
//		xfranc@dial.oleane.com , xfra@balise.com
//
//-----------------------------------------------------------------------

#ifndef WINDOWS
# include <unistd.h>
# include <sys/types.h>
# include <sys/mman.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <errno.h>

#else

# include <windows.h>
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <errno.h>

#endif

#include "filemap.h"

TStatus	fileMap (const char *path, const char *&address_, int &size, long &sysctx)
{
#ifndef WINDOWS
	int fd = open (path, O_RDONLY);
	if (fd < 0)
		return "no access";
	struct stat	statd;
	if (fstat(fd, & statd) < 0)
		return "bad size";
	if (!S_ISREG(statd.st_mode))
		return "not a regular file";
	size = statd.st_size;
	address_ = (const char*) mmap (nil, size, PROT_READ, MAP_PRIVATE, fd, 0);
	if (address_ == (const char*)-1)
		return "cannot map";
	sysctx = (long) fd;
	return nil;	// OK

#else
	HANDLE	fileHdl = CreateFile (path, GENERIC_READ, FILE_SHARE_READ, NULL,
				      OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (fileHdl == INVALID_HANDLE_VALUE)
		return "no access";

	DWORD sizeHigh, ssize = GetFileSize(fileHdl, &sizeHigh);
	if (ssize <= 0)
		return "bad size";
 
	HANDLE mapHdl = CreateFileMapping(fileHdl, NULL, PAGE_READONLY, 0, 0, NULL);
	if (mapHdl == NULL)
		return "cannot create mapping";
	CloseHandle(fileHdl);	// it's OK for this guy

	address_ = (const char*) MapViewOfFile (mapHdl, FILE_MAP_READ, 0, 0, 0);
	if (address_ == NULL) {
		CloseHandle(mapHdl);
		return "cannot map";
	}
	size = ssize;
	sysctx = (long) mapHdl;
	return nil;	// OK
#endif
}


TStatus	fileUnmap (const char *address_, int size, long sysctx)
{
	if (address_ != nil)
	{
#ifndef WINDOWS
		munmap ((caddr_t) address_, size);
		close((int) sysctx);
#else
		UnmapViewOfFile(address_);
		CloseHandle((HANDLE) sysctx);
#endif
		address_ = nil;
	}
	return nil;
}

