//-----------------------------------------------------------------------
//
//	Project TibThob: a tool to search the Rangjung Yeshe Dictionary
//
//
//	This program is subject to the GNU GENERAL PUBLIC LICENSE
//	It is provided as is - WITHOUT ANY WARRANTY - in the hope
//	it will be of benefit to all beings, or at least some of them.
//	
//-----------------------------------------------------------------------
//	Copyright Xavier C. FRANC 1999
//	report bugs and comments at:
//		xfranc@dial.oleane.com , xfra@balise.com
//
//-----------------------------------------------------------------------

#include "hashtbl.h"

HashTable::HashTable(int size)
{
	size_ = 1;
	while(size_ < size)
		size_ *= 2;
	arena_ = new Link* [size_];
	for(int s = size_; --s >= 0; )
		arena_[s] = nil;
}

HashTable::~HashTable()
{
	for(int s = size_; --s >= 0; ) {
		for(Link *nxt, *lnk = arena_[s]; lnk != nil; lnk = nxt) {
			nxt = lnk->next;
			delete [] lnk->key;
			delete lnk;
		}
	}
	delete [] arena_;
}

int	HashTable::hash (const char *str, int length)
{
	int h = 0;
	for (const char *p = str; *p != 0 && --length >= 0; p++)
		h = (h << 1) + *p;
	return h;
}

void	*HashTable::find (int hash, const char *string, int length)
{
	if (length < 0)
		length = strlen(string);
	for (Link *lnk = arena_[hash & (size_ - 1)]; lnk != nil; lnk = lnk->next)
		if (strncmp(lnk->key, string, length) == 0 && lnk->key[length] == 0)
			return lnk->assoc;
	return nil; // if assoc == nil, we dont see the difference...
}

void	*HashTable::find (const char *string, int length)
{
	if (length < 0)
		length = strlen(string);
	return find(hash(string, length), string, length);
}

void	HashTable::insert (void *data, int hash, const char *string, int length)
{
	if (length < 0)
		length = strlen(string);
	int slot = hash & (size_ - 1);
	for (Link *lnk = arena_[slot]; lnk != nil; lnk = lnk->next)
		if (strncmp(lnk->key, string, length) == 0 && lnk->key[length] == 0) {
			lnk->assoc = data;
			return;
		}
	Link *newLnk = new Link;
	newLnk->assoc = data;
	newLnk->next = arena_[slot];
	arena_[slot] = newLnk;
	newLnk->key = new char [length + 1];
	memcpy(newLnk->key, string, length);
	newLnk->key[length] = 0;
}

void	HashTable::insert (void *data, const char *string, int length)
{
	if (length < 0)
		length = strlen(string);
	insert(data, hash(string, length), string, length);
}
