//-----------------------------------------------------------------------
//
//	Project TibThob: a tool to search the Rangjung Yeshe Dictionary
//
//
//	This program is subject to the GNU GENERAL PUBLIC LICENSE
//	It is provided as is - WITHOUT ANY WARRANTY - in the hope
//	it will be of benefit to all beings, or at least some of them.
//	
//-----------------------------------------------------------------------
//	Copyright Xavier C. FRANC 1999
//	report bugs and comments at:
//		xfranc@dial.oleane.com , xfra@balise.com
//
//-----------------------------------------------------------------------

#include <tk.h>
#include "tsrcheng.h"
#include <time.h>

// one global engine (poor practice, I know)
TibSearchEngine *TheEngine = nil;


static int	Error (Tcl_Interp *interp, char *text)
{
	Tcl_AppendResult(interp, text, NULL);
	return TCL_ERROR;
}

static int	numResult (Tcl_Interp *interp, int value)
{
	char num[20];
	sprintf(num, "%d", value);
	Tcl_AppendResult(interp, num, NULL);
	return TCL_OK;
}

static TibSearchEngine::Restrict cvArea (const char *area)
{
	switch(area[0]) {
	case 'S':
		return TibSearchEngine::AtStart;
	case 'E':
		return TibSearchEngine::InEntry;
	case 'D':
		return TibSearchEngine::InDefinition;
	default:
		return TibSearchEngine::Anywhere;
	}
}

//--------------------------------------------------------------------------------
//	Displaying the results: we call back Tk to write into the text window
//

#define DISPLAY_WY_E "$txtwin insert end $txt WylieEntry"
#define DISPLAY_WY_HI "$txtwin insert end $txt WylieHigh"
#define DISPLAY_EN_D "$txtwin insert end $txt Definition"
#define DISPLAY_EN_HI "$txtwin insert end $txt DefHigh"
Tcl_Obj	*Cmd_WY_E = nil, *Cmd_WY_HI, *Cmd_EN_D, *Cmd_EN_HI;

static
void	displayChunk (Tcl_Interp *interp, Tcl_Obj *cmd, const char *str, int len)
{
	const int MaxSize = 10000;
	char	tmpbuf[MaxSize+1];
	if (len < 0)
		Tcl_SetVar(interp, "txt", (char*) str, 0);
	else {
		if (len > MaxSize) len = MaxSize;	// brutal...
		memcpy(tmpbuf, str, len);
		tmpbuf[len] = 0;
		Tcl_SetVar(interp, "txt", tmpbuf, 0);
	}
	Tcl_EvalObj(interp, cmd);
}

extern	const char	*ciSearchString (const char *start, const char *end,
				       const char *string, int length);

//	displays a string with hits highlighted
static
void	displayHigh (Tcl_Interp *interp, const char *str, const char *searched,
		     Tcl_Obj *normal, Tcl_Obj *high)
{
	int	srchlen = strlen(searched);
	const char *hit = str, *last = str, *end = str + strlen(str);
	for ( ; (hit = ciSearchString(last, end, searched, srchlen)) != nil; ) {
		if (hit > last)
			displayChunk(interp, normal, last, hit - last);
		displayChunk(interp, high, hit, srchlen);
		last = hit + srchlen;
	}
	if (end > last)
		displayChunk(interp, normal, last, end - last);
}

static	void	prepareTags(int init)
{
    if (init) {
	Tcl_IncrRefCount(Cmd_WY_E = Tcl_NewStringObj(DISPLAY_WY_E, -1));
	Tcl_IncrRefCount(Cmd_WY_HI = Tcl_NewStringObj(DISPLAY_WY_HI, -1));
	Tcl_IncrRefCount(Cmd_EN_D = Tcl_NewStringObj(DISPLAY_EN_D, -1));
	Tcl_IncrRefCount(Cmd_EN_HI = Tcl_NewStringObj(DISPLAY_EN_HI, -1));
    }
    else {
	Tcl_DecrRefCount(Cmd_WY_E);
	Tcl_DecrRefCount(Cmd_WY_HI);
	Tcl_DecrRefCount(Cmd_EN_D);
	Tcl_DecrRefCount(Cmd_EN_HI);
    }
}

static
void	doDisplay (Tcl_Interp *interp, int highlight, char * /*tibFont*/)
{
	const char *entry, *def, *searched = TheEngine->searched();
	TheEngine->firstHit ();
	while (TheEngine->getNextHit(entry, def)) {
		//printf("%s :\n\t%s\n", entry, def);
		if (!highlight)
			displayChunk (interp, Cmd_WY_E, entry, -1);
		else	displayHigh (interp, entry, searched, Cmd_WY_E, Cmd_WY_HI);
		displayChunk (interp, Cmd_WY_E, "\n", 1);

		if (!highlight)
			displayChunk (interp, Cmd_EN_D, def, -1);
		else	displayHigh (interp, def, searched, Cmd_EN_D, Cmd_EN_HI);
		displayChunk (interp, Cmd_WY_E, "\n", 1);
	}
}

//	Display for phonetics and approx Wylie:
static
void	doConvDisplay (Tcl_Interp *interp, int highlight)
{
	const char *entry, *def;
	int	pos, size;
	TheEngine->firstHit ();
	while (TheEngine->getNextConvHit(entry, pos, size, def))
	{
		if (!highlight)
			displayChunk (interp, Cmd_WY_E, entry, -1);
		else {
			displayChunk (interp, Cmd_WY_E, entry, pos);
			displayChunk (interp, Cmd_WY_HI, entry + pos, size);
			displayChunk (interp, Cmd_WY_E, entry + pos + size, -1);
		}
		displayChunk (interp, Cmd_WY_E, "\n", 1);

		displayChunk (interp, Cmd_EN_D, def, -1);
		displayChunk (interp, Cmd_WY_E, "\n", 1);
	}
}

//--------------------------------------------------------------------------------
//	Exporting

#if FUTURE_FEATURE
enum StyleItem { STpre, STpost, STePre, STePost, STdPre, STdPost };
static char *const headers[] = {
	"PRE ", "POST ", "EPRE ", "EPOST ", "DPRE ", "DPOST "
};

TStatus	loadExportStyle (char *path, char *req, char *styles[])
{
	int i, ok = false;
	for(i = STpre; i <= STdPost; i++)
		styles[i] = nil;
	FILE *in = fopen(path, "r");
	if (in == NULL)
		return "access";
	char buf[1000];
	for(; fgets(buf, 1000, in); ) {
		if (buf[0] == '#')
			continue;
		if (strncmp(buf, "EXPORT ", 7) == 0)
			ok = strncmp(buf + 7, req, strlen(req)) == 0;
		else if (!ok)
			continue;
		else for(i = STpre; i <= STdPost; i++)
			if(strncmp(buf, headers[i], strlen(headers[i])) == 0) {
				styles[i] = getItem(buf + strlen(headers[i]));
				
			}
	}
	fclose(in);
	return nil;
}
#endif

static
int	doExport (Tcl_Interp *interp, char *style, char *file, char */*fstyles*/)
{
	const char *entry, *def;
	FILE *out = fopen(file, "w");
	if (out == NULL)
		return Error(interp, "cannot write");
	TheEngine->firstHit ();
	while (TheEngine->getNextHit(entry, def)) {
		// TODO: styles
		fprintf(out, "%s\t%s\n", entry, def);
	}
	fclose(out);
	return TCL_OK;
}

//--------------------------------------------------------------------------------
//	Tcl interface function with Tibetan Search Engine
//

#ifndef NO_MKTCLAPP
#include "../appinit.h"
extern "C" int	ET_COMMAND_tibdix(ET_TCLARGS) {
#else
extern "C" int     App_tibdix (ClientData /*clientData*/, Tcl_Interp *interp,
			       int argc, char * argv[]) {
#endif
	if (argc < 2) {
		Tcl_AppendResult(interp, "bad argument count", NULL);
		return TCL_ERROR;
	}
	if (TheEngine == nil)
		TheEngine = new TibSearchEngine;
	long t = clock();
	if (streq(argv[1], "dictionary")) {
		// initialise dictionary: <dictionary_path> 
		TStatus st = TheEngine->open(argv[2]);
		if (st == nil)
			return TCL_OK;
		return Error(interp, st);
	}
	else if (streq(argv[1], "hitmax")) {
		int max = 0;
		if (argc >= 3)
			max = atoi(argv[2]);
		TheEngine->setHitMax(max);
	}
	else if (streq(argv[1], "require")) {
		// add a require specification: string + area
		if (argc < 4)
			return Error (interp, "require: bad argc");
		TheEngine->require(argv[2], cvArea(argv[3]));
	}
	else if (streq(argv[1], "exclude")) {
		// add an exclude specification: string + area
		if (argc < 4)
			return Error (interp, "exclude: bad argc");
		TheEngine->exclude(argv[2], cvArea(argv[3]));
	}
	else if (streq(argv[1], "search")) {
		// args: <string> <area> 
		// returns the hit count
		if (argc < 4)
			return Error (interp, "search: bad argc");
		numResult(interp, TheEngine->search(argv[2], cvArea(argv[3])));
	}
	else if (streq(argv[1], "phonsearch")) {
	    // args: <string> <area> 
	    // returns the hit count
	    if (argc < 4)
		    return Error (interp, "phonsearch: bad argc");
	    numResult(interp, TheEngine->searchPhonetics(argv[2], cvArea(argv[3])));
	}
	else if (streq(argv[1], "approxsearch")) {
	    // args: <string> <area> 
	    // returns the hit count
	    if (argc < 4)
		    return Error (interp, "approxsearch: bad argc");
	    numResult(interp, TheEngine->searchApproxWylie(argv[2], cvArea(argv[3])));
	}
	else if (streq(argv[1], "phonetics")) {
		if (argc < 3)
			return Error (interp, "phonetics: bad argc");
		TStatus st = TheEngine->initPhonetics(argv[2]);
		if (st != NULL)
			return Error(interp, st);
	}
	else if (streq(argv[1], "approx")) {
		if (argc < 3)
			return Error (interp, "approx Wylie: bad argc");
		TStatus st = TheEngine->initApproxWylie(argv[2]);
		if (st != NULL)
			return Error(interp, st);
	}
	else if (streq(argv[1], "display")) {
		// args: <view> <highlight>? <font>?
		if (argc < 3)
			return Error (interp, "display: bad argc");
		Tcl_SetVar(interp, "txtwin", argv[2], 0);
		int highlight = false;
		char *font = nil;
		if (argc >= 4) {
			if(argv[3][0] == '1')
				highlight = true;
			if (argc >= 5)
				font = argv[4];
		}
		prepareTags(true);
		if (TheEngine->lastSearch() != 'W')
			doConvDisplay(interp, highlight);
		else	doDisplay(interp, highlight, font);
		prepareTags(false);
	}
	else if (streq(argv[1], "export")) {
		// args: <style> <file>
		if (argc < 4)
			return Error (interp, "export: bad argc");
		return doExport(interp, argv[2], argv[3], argv[4]);
	}
	else if (streq(argv[1], "reset")) {
		if (!TheEngine->reset())
			return Error(interp, "dictionary non initialised");
	}
	else {
		return Error (interp, "bad command");
	}
	//printf("time for %s: %d\n", argv[1], clock() - t);
	return TCL_OK;
}

#ifdef NO_MKTCLAPP
//	plug-in point:
//
extern "C" int     _DLL_EXPORT Tibthob_Init (Tcl_Interp *interp)
{
        Tcl_CreateCommand (interp, "tibdix", App_tibdix, 0, NULL);
	return TCL_OK;
}
#endif
