//-----------------------------------------------------------------------
//
//	Project TibThob: a tool to search the Rangjung Yeshe Dictionary
//
//
//	This program is subject to the GNU GENERAL PUBLIC LICENSE
//	It is provided as is - WITHOUT ANY WARRANTY - in the hope
//	it will be of benefit to all beings, or at least some of them.
//	
//-----------------------------------------------------------------------
//	Copyright Xavier C. FRANC 1999
//	report bugs and comments at:
//		xfranc@dial.oleane.com , xfra@balise.com
//
//-----------------------------------------------------------------------

#include "tsrcheng.h"
#include "filemap.h"
#include "wyliphon.h"
#include "wylieapp.h"
#include "tibphon.h"
#include "hashtbl.h"

const char	*searchString ( const char *start, const char *limit,
			        const char *string, int length )
{
    char  first = string[0];
    char  second = string[1];
    string += 2;
    length -= 2;
    while (true) {
	const char *nxt = (char*) memchr(start, first, limit - start);
	if (nxt == nil)
	    return nil;
	start = nxt + 1;
	if (second == 0 ||
	    second == *start && memcmp(start+1, string, length) == 0)
	    return nxt;
    }
}

#define upCase(c) UpCaseTable[ ((int) (c)) & 0xff ]
static char UpCaseTable[] = {
' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 
' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 
' ', '!', '"', '#', '$', '%', '&', '\'', '(', ')', '*', '+', ',', '-', '.', '/', 
'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ':', ';', '<', '=', '>', '?', 
'@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 
'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^', '_', 
'`', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 
'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '{', '|', '}', '~', ' ', 
' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 
' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 
'', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', 
'', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', 
'', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', 
'', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', 
'', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', 
'', '', '', '', '', '', '', '', '', '', '', '', '', '', '', ' '
};

const char	*ciSearchString( const char *start, const char *limit,
				 const char *str, int length )
{
    if (length < 0)
	length = strlen(str);
    // string to search must be case folded
    StringList cleaner;
    char *copy = cleaner.chars_ = (char*) malloc(length + 1);
    for(int i = 0; i < length; i++)
	copy[i] = upCase(str[i]);
    copy[length] = 0;

    char first = *copy++;
    -- length;
    while (true) {
	const char *nxt = start;
	for ( ; nxt < limit; nxt++)
	    if (upCase(*nxt) == first)
		break;
	if (nxt >= limit)
	    return nil;
	int i = 0;
	start = nxt + 1;
	for( ; i < length; i++)
	    if (upCase(start[i]) != copy[i])
		break;
	if (i >= length) {
	    return nxt;
	}
	start = nxt + 1;
    }
}


TibSearchEngine::TibSearchEngine()
{
	dict_ = nil;
	dictSize_ = 0;

	searched_ = nil;
	searchedLen_ = 0;
	w2p_ = nil;
	wapp_ = nil;

	requires_ = excludes_ = nil;
	hitEnts_ = hitDefs_ = nil;
	curHitEnt_ = curHitDef_ = nil;
	hitMax_ = 10000000;	// the sky is he limit
}

TibSearchEngine::~TibSearchEngine()
{
	reset();
	delete w2p_;
}

TStatus		TibSearchEngine::open (char *path)
{
	if (dict_ != nil) {
		fileUnmap(dict_, dictSize_, mapping_);
		dict_ = nil;
	}
	if (path == nil)	// just close
		return nil;	
	TStatus	st = fileMap(path, dict_, dictSize_, mapping_);
	return st;
}

int		TibSearchEngine::reset()
{
	restart();
	clearList(requires_);
	clearList(excludes_);
	return dict_ != nil;
}

void		TibSearchEngine::restart()
{
	clearList(hitEnts_);
	clearList(hitDefs_);
	curHitEnt_ = curHitDef_ = nil;
}

char		*trimAndCopy (const char *start, int length)
{
	while(length > 0 && isspace(*start))
		start++, length--;
	while(length > 0 && isspace(start[length-1]))
		length--;
	char *copy = (char*) malloc(length + 1);
	memcpy(copy, start, length);
	copy[length] = 0;
	return copy;
}

int		TibSearchEngine::search (char *string, Restrict mode)
{
    int	hits = 0;
    const char *dictEnd = dict_ + dictSize_;

    // must copy: else it will be destroyed later by replace()...
    string = strdup(string);
    if (strlen(string) == 0)	// dont want so many hits...
	return -1;
    replace(searched_, string);
    searchedLen_ = strlen(string);
    lastSearch_ = 'W';

    const char *end = dict_;
    while (true) {
	const char *hit = ciSearchString(end, dictEnd, searched_, searchedLen_);
	if (hit == nil)
	    break;
	end = searchString(hit, dictEnd, "\n", 1);
	if (end == nil)
	    end = dictEnd;
	if (realHit (hit, end, mode, hits))
	    hits ++;
	if (++ end >= dictEnd)
	    break;
    }
    curHitEnt_ = hitEnts_;
    curHitDef_ = hitDefs_;
    return hits;
}

int		TibSearchEngine::searchPhonetics (char *string, Restrict mode)
{
    if (w2p_ == nil)
	return -999;	// not initialised
    int st = phoneticString(string);
    if (st < 0)
	return st;	// improper input: -error_position-1
    int	hits = 0;
    const char *dictEnd = dict_ + dictSize_;

    const char *cur = dict_, *sep;
    while (cur < dictEnd) {
	int pos = phoneticMatch(cur, sep);
	sep = searchString(sep, dictEnd, "\n", 1);
	if (sep == nil)
	    sep = dictEnd;
	else if (pos > 0 && mode == AtStart)
	    pos = -1;
	if (pos >= 0 && realHit (cur, sep, mode, hits)) {
	    hits ++;
	}
	cur = sep + 1;
    }
    curHitEnt_ = hitEnts_;
    curHitDef_ = hitDefs_;
    return hits;
}

int		TibSearchEngine::searchApproxWylie (char *string, Restrict mode)
{
    if (wapp_ == nil)
	return -999;	// not initialised
    int st = approxWylie(string);
    if (st < 0)
	return st;	// improper input: -error_position-1
    int	hits = 0;
    const char *dictEnd = dict_ + dictSize_;

    const char *cur = dict_, *sep;
    while (cur < dictEnd) {
	int pos = approxWylieMatch(cur, sep);
	sep = searchString(sep, dictEnd, "\n", 1);
	if (sep == nil)
	    sep = dictEnd;
	else if (pos > 0 && mode == AtStart)
	    pos = -1;
	if (pos >= 0 && realHit (cur, sep, mode, hits))
	    hits ++;
	cur = sep + 1;
    }
    curHitEnt_ = hitEnts_;
    curHitDef_ = hitDefs_;
    return hits;
}

int		TibSearchEngine::realHit (const char *hit, const char *end,
					  Restrict restrict, int hitCount)
{
	const char *start = hit;
	for(; start > dict_ && start[-1] != '\n'; )
		start --;
	const char *sep = searchString(start, end, "-", 1);
	if (sep == nil)
		// malformed entry: could be simple text
		restrict = Anywhere, sep = start;
	switch(restrict) {
	case Anywhere:
		break;
	case AtStart:
		if (hit != start)
			return false;
		break;
	case InEntry:
		if (hit >= sep)
			return false;
		break;
	case InDefinition:
		if (hit <= sep)
			return false;
		break;
	}
	//printf("HIT at %d ", start); fwrite(start, 1, end + 1 - start, stdout);

	// check the requires:
	StringList *check;
	for (check = requires_ ; check != nil; check = check->next_) {
		char *string = check->chars_ + 1;
		switch(check->chars_[0]) {
		case 'S':	// at start
			if (strncmp(start, string, strlen(string)) != 0)
				return false;
			break;
		case 'E':	// in entry
			if (ciSearchString(start, sep, string, strlen(string)) == nil)
				return false;
			break;
		case 'D':	// in definition
			if (ciSearchString(sep, end, string, strlen(string)) == nil)
				return false;
			break;
		default:	// anywhere
			if (ciSearchString(start, end, string, strlen(string)) == nil)
				return false;
			break;
		}
	}
	// check the excludes:
	for (check = excludes_ ; check != nil; check = check->next_) {
		char *string = check->chars_ + 1;
		switch(check->chars_[0]) {
		case 'S':	// at start
			if (strncasecmp(start, string, strlen(string)) == 0)
				return false;
			break;
		case 'E':
			if (ciSearchString(start, sep, string, strlen(string)) != nil)
				return false;
			break;
		case 'D':
			if (ciSearchString(sep, end, string, strlen(string)) != nil)
				return false;
			break;
		default:
			if (ciSearchString(start, end, string, strlen(string)) != nil)
				return false;
			break;
		}
	}

	if (hitCount >= hitMax_)
	    return true;
	// store
	StringList *SL = new StringList;
	SL->chars_ = trimAndCopy(start, sep - start);
	if (curHitEnt_ == nil)
		hitEnts_ = SL;
	else	curHitEnt_->next_ = SL;
	curHitEnt_ = SL;
	SL = new StringList;
	++ sep;
	SL->chars_ = trimAndCopy(sep, end - sep);
	if (curHitDef_ == nil)
		hitDefs_ = SL;
	else	curHitDef_->next_ = SL;
	curHitDef_ = SL;
	return true;
}

//--------------------------------------------------------------------------------
//

int		TibSearchEngine::phoneticMatch (const char *start, const char * &sep)
{
	char	phonemes[1000];

	int	phl = phonConvert (start, sep, phonemes);
	// all conversion has been done by phoneConvert()
	// we search the string of phonetic symbols like a normal string:
	const char *hit = searchString(phonemes, phonemes + phl,
				       searched_, searchedLen_);
	return hit == nil ? -1 : (hit - phonemes);
}

int	TibSearchEngine::phonConvert (const char *buffer, const char * &sep,
				      char phonemes[])
{
	int pos = 0, php = 0;
	for(;;) {
		while(buffer[pos] == ' ' || buffer[pos] == ',') pos++;
		char c = buffer[pos];
		if (c == 0 || c == '-' || c == '}' || c == '\r' || c == '\n')
			break;
		int size = w2p_->parse(buffer + pos);
		if (size == 0)
			break;
		int phl = w2p_->convert(phonemes + php);
		if (phl > 0)
			php += phl;
		else {
#ifdef DEBUG
			char erron[100];
			memcpy(erron, buffer + pos, size); erron[size] = 0;
			fprintf(stderr, "cannot convert '%s'\n", erron);
#endif
		}
		pos += size;
	}
	sep = buffer + pos;
	phonemes[php] = ph_void;
	return php;
}

int	TibSearchEngine::phoneticString (const char *string)
{
	if (string[0] == 0)
		return -1;
	char	phonemes[1000];
	int	phsL = w2p_->input2Phonemes (string, phonemes);
	if (phsL < 0)
		return phsL;
	char *phs = new char [phsL+1];
	memcpy(phs, phonemes, phsL);
	phs[phsL] = 0;
	replace(searched_, phs);
	searchedLen_ = phsL;
	lastSearch_ = 'P';
#if 1
	char buf[1000];
	extern void	convertPhonemes (const char *phons, char *buffer);
	convertPhonemes(searched_, buf); printf("-> %s\n", buf);
#endif
	return 0;
}


TStatus		TibSearchEngine::initPhonetics (const char *tablePath)
{
	Wylie2Phonetics *w = new Wylie2Phonetics();
	TStatus st = w->loadTable (tablePath);
	if (st != nil) {
		delete w;
		return st;
	}
	delete w2p_;
	w2p_ = w;
	return nil;
}


TStatus		TibSearchEngine::initApproxWylie (const char *tablePath)
{
	WylieReducer *w = new WylieReducer();
	TStatus st = w->loadTable (tablePath);
	if (st != nil) {
		delete w;
		return st;
	}
	delete wapp_;
	wapp_ = w;
	return nil;
}

int		TibSearchEngine::approxWylieMatch (const char *start, const char * &sep)
{
	char	equiv[1000];

	int	eqL = approxWylieConvert (start, sep, equiv);
	// all conversion has been done by phoneConvert()
	// we search the string of phonetic symbols like a normal string:
	const char *hit = searchString(equiv, equiv + eqL,
				       searched_, searchedLen_);
	return hit == nil ? -1 : (hit - equiv);
}

int	TibSearchEngine::approxWylie (const char *string)
{
	if (string[0] == 0)
		return -1;
	char	equiv[1000];
	const char *e;
	int	eqL = approxWylieConvert(string, e, equiv);
	if (eqL <= 0)
		return eqL-1;
	eqL --;	// remove final space
	char *phs = new char [eqL+1];
	memcpy(phs, equiv, eqL);
	phs[eqL] = 0;
	replace(searched_, phs);
	printf("approxWylie |%s|\n", phs);
	searchedLen_ = eqL;
	lastSearch_ = 'A';
	return 0;
}

int	TibSearchEngine::approxWylieConvert (const char *buffer, const char * &sep,
					     char equiv[])
{
	int pos = 0, php = 0;
	for(;;) {
		while(buffer[pos] == ' ' || buffer[pos] == ',') pos++;
		char c = buffer[pos];
		if (c == 0 || c == '-' || c == '}' || c == '\r' || c == '\n')
			break;
		int size = wapp_->parse(buffer + pos);
		if (size == 0)
			break;
		int phl = wapp_->convert(equiv + php);
		if (phl > 0)
			php += phl;
		else {
#ifdef DEBUG
			char erron[100];
			memcpy(erron, buffer + pos, size); erron[size] = 0;
			fprintf(stderr, "app. Wylie untranslated '%s'\n", erron);
#endif
		}
		pos += size;
	}
	sep = buffer + pos;
	equiv[php] = 0;
	return php;
}



char	RestrictCode[] = { 'A', 'S', 'E', 'D' };

void		TibSearchEngine::require (char *string, Restrict restrict)
{
	int L = strlen(string);
	if (L == 0)
		return;
	char *stored = (char*) malloc(L + 2);
	// we code the restrict as first char...
	stored[0] = RestrictCode[restrict];
	strcpy(stored + 1, string);
	StringList *SL = new StringList;
	SL->chars_ = stored;
	SL->next_ = requires_;
	requires_ = SL;
}

void		TibSearchEngine::exclude (char *string, Restrict restrict)
{
	int L = strlen(string);
	if (L == 0)
		return;
	char *stored = (char*) malloc(L + 2);
	// we code the restrict as first char...
	stored[0] = RestrictCode[restrict];
	strcpy(stored + 1, string);
	StringList *SL = new StringList;
	SL->chars_ = stored;
	SL->next_ = excludes_;
	excludes_ = SL;
}

void	TibSearchEngine::setHitMax (int max)
{
	hitMax_ = max;
}

void	TibSearchEngine::firstHit ()
{
	curHitEnt_ = hitEnts_;
	curHitDef_ = hitDefs_;
}

int	TibSearchEngine::getNextHit (const char * &entry, const char * &def)
{
	if (curHitEnt_ == nil)
		return false;
	entry = curHitEnt_->chars_;
	def = curHitDef_->chars_;
	curHitEnt_ = curHitEnt_->next_;
	curHitDef_ = curHitDef_->next_;
	return true ;
}

int	TibSearchEngine::getNextConvHit (const char * &entry, int &wpos, int &wsize,
					 const char * &def)
{
	if (curHitEnt_ == nil)
		return false;
	entry = curHitEnt_->chars_;
	def = curHitDef_->chars_;
	curHitEnt_ = curHitEnt_->next_;
	curHitDef_ = curHitDef_->next_;
	WylieParser *parser = (lastSearch_ == 'P') ? w2p_ : (WylieParser*) wapp_;

	// locate the hit:
	char	equiv[1000];
	int	pos = 0, eqp = 0, positions[1000], ppp = 0;
	//fprintf(stderr, "ENTRY: %s\n", entry);
	for(;;) {
		while(entry[pos] == ' ' || entry[pos] == ',') pos++;
		char c = entry[pos];
		if (c == 0 || c == '-' || c == '}' || c == '\r' || c == '\n')
			break;
		int size = parser->parse(entry + pos);
		if (size == 0)
			break;
		// remember the pair (wylie_pos, phonetic_pos):
		positions[ppp++] = pos;
		positions[ppp++] = eqp;
		// 
		int phl = parser->convert(equiv + eqp);
		if (phl > 0)
			eqp += phl;
		pos += size;
	}
	equiv[eqp] = ph_void;
	positions[ppp++] = pos;
	positions[ppp++] = eqp;

	const char *hit = searchString(equiv, equiv + eqp,
				       searched_, searchedLen_);

	// we have position and size in phonetics: convert to pos and size in Wylie:
	wpos = 0; wsize = 0;
	if (hit != nil) {	// should not happen
		int phonPos = hit - equiv;
		int p = 0;
		//for(int q = 0; q < ppp; q += 2) fprintf(stderr, " (%d, %d)", positions[q], positions[q+1]); fprintf(stderr, "\n");
		for(; p < ppp; p += 2) {
			wpos = positions[p];
			if (positions[p+1] >= phonPos)
				break;
		}
		for(; p < ppp; p += 2) {
			wsize = positions[p] - wpos;
			if (positions[p+1] >= phonPos + searchedLen_)
				break;
		}
	}
	return true ;
}

void	TibSearchEngine::clearList (StringList *&list)
{
	StringList *L, *N;
	for (L = list; L != nil; L = N) {
		N = L->next_;
		delete L;
	}
	list = nil;
}
