#ifndef _Included_wyliparse_h
#define _Included_wyliparse_h

//-----------------------------------------------------------------------
//
//	Project TibThob: a tool to search the Rangjung Yeshe Dictionary
//
//
//	This program is subject to the GNU GENERAL PUBLIC LICENSE
//	It is provided as is - WITHOUT ANY WARRANTY - in the hope
//	it will be of benefit to all beings, or at least some of them.
//	
//-----------------------------------------------------------------------
//	Copyright Xavier C. FRANC 1999
//	report bugs and comments at:
//		xfranc@dial.oleane.com , xfra@balise.com
//
//-----------------------------------------------------------------------

#include "tibbasic.h"
class	HashTable;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//	class WylieParser : 
//	Generic Wylie Syllable parser: works by simply splitting a syllable
//	in three parts:
//	- vowel of central letter {ming gzhi}
//	- left-part: including ming gzhi, prefix (ngonjuk), superscript, subjoined
//	- right-part: including suffix, post-suffix, achung with vowel
//	
//	It can be equipped with one associative table for the left-part, and
//	one for the right-part, for specialised purpose like phonetic or other
//	translations.

class	WylieParser
{
// .......... public methods ..................................
public:
	WylieParser();
virtual	~WylieParser();

virtual	int	convert (char equiv[]);

	//------- feeding the associative tables
virtual	void	leftAssoc (char *leftPart, void *data);
virtual	void	rightAssoc (char *rightPart, void *data);

	// parse a syllable: returns its length. Stops on any non-Wylie char
	// including the space (considered as syllable separator).
	// 'wylie' is assumed to point the start of a syllable (no leading space)
	// Associated data is retrieved by the following functions.
virtual	int	parse (const char *wylie, int length = -1);

	// get left part: if a non-null buffer is provided, copy left chars to it
	// returns the data associated by leftAssoc() (see below)
	// [NB: The 'wylie' buffer must still be valid]
virtual	void	*getLeft(char *buffer = nil);
	// get right part: 
virtual	void	*getRight(char *buffer = nil);
	// get central vowel
virtual	char	getVowel();

// .......... internal methods ................................
protected:

// .......... instance variables ..............................
protected:
	// buffer given by parse(): must remain valid
	const char 	*wylie_;
	// size of left part
	int	leftSize_;
	// hash value of left part (computed by parse())
	int	leftHash_;
	// position and size of right part
	int	rightPos_;
	int	rightSize_;
	// hash value of right part (computed by parse())
	int	rightHash_;

	HashTable *leftTable_;
	HashTable *rightTable_;

}; // end of class WylieParser

#endif
// DO NOT INSERT after this #endif 
