//-----------------------------------------------------------------------
//
//	Project TibThob: a tool to search the Rangjung Yeshe Dictionary
//
//
//	This program is subject to the GNU GENERAL PUBLIC LICENSE
//	It is provided as is - WITHOUT ANY WARRANTY - in the hope
//	it will be of benefit to all beings, or at least some of them.
//	
//-----------------------------------------------------------------------
//	Copyright Xavier C. FRANC 1999
//	report bugs and comments at:
//		xfranc@dial.oleane.com , xfra@balise.com
//
//-----------------------------------------------------------------------

#include "wylipars.h"
#include "hashtbl.h"

WylieParser::WylieParser()
{
	leftTable_ = new HashTable(900);
	rightTable_ = new HashTable(60);
}

WylieParser::~WylieParser()
{
	delete leftTable_;
	delete rightTable_;
}

int	WylieParser::convert (char [])
{
	return -1;
}

void	WylieParser::leftAssoc (char *left, void *data)
{
	if (leftTable_->find(left))
		fprintf(stderr, "Warning: duplicate left-assoc '%s'\n", left);
	leftTable_->insert(data, HashTable::hash(left, strlen(left)), left);
}

void	WylieParser::rightAssoc (char *right, void *data)
{
	if (rightTable_->find(right))
		fprintf(stderr, "Warning: duplicate right-assoc '%s'\n", right);
	rightTable_->insert(data, HashTable::hash(right, strlen(right)), right);
}

const char	*Vowels = "aeiouAEIOU";
const char	*Separators = "-!?|{}()[]<>/;:,";

int	WylieParser::parse (const char *wylie, int length)
{
	if (length < 0)
		length = 10; ///strlen(wylie);
	wylie_ = wylie;
	leftHash_ = 0;
	rightSize_ = 0;
	rightHash_ = 0;
	const char *p = wylie, *end = wylie + length;

	// scan left part
	for( ; p < end; p++) {
		if (strchr(Vowels, *p) != nil)
			break;
		leftHash_ = (leftHash_ << 1) + *p;
	}
	leftSize_ = p - wylie;	// null length means big A

	if (p >= end)
		return leftSize_;	// should not happen

	p++;	// skip vowel
	for( ; p < end; p++) {
		if (*p <= ' ' || strchr(Separators, *p) != nil)
			break;
		rightHash_ = (rightHash_ << 1) + *p;
	}
	rightSize_ = (p - wylie) - leftSize_ - 1;
	return p - wylie;
}

void	*WylieParser::getLeft (char *buffer)
{
	if (buffer != nil) {
		memcpy(buffer, wylie_, leftSize_);
		buffer[leftSize_] = 0;
	}
	return leftTable_->find(leftHash_, wylie_, leftSize_);
}

void	*WylieParser::getRight (char *buffer)
{
	if (buffer != nil) {
		memcpy(buffer, wylie_ + leftSize_ + 1, rightSize_);
		buffer[rightSize_] = 0;
	}
	return rightTable_->find(rightHash_, wylie_ + leftSize_ + 1, rightSize_);
}

char	WylieParser::getVowel()
{
	return wylie_[leftSize_];
}
